import numpy as np
from scipy.linalg import eigh


class DeformationMethod:
    def __init__(
        self,
        num_input,
        E_inp=1,
        I_inp=1,
        A_inp=1,
        rho_inp=1,
        l_inp=1,
        static_inp="SPG",
        l_inp_multi=[1, 1],
        _num_integral = 1000
    ):
        """_summary_

        Args:
            num_input (_type_): _description_
            E_inp (int, optional): _description_. Defaults to 1.
            I_inp (int, optional): _description_. Defaults to 1.
            A_inp (int, optional): _description_. Defaults to 1.
            rho_inp (int, optional): _description_. Defaults to 1.
            l_inp (int, optional): _description_. Defaults to 1.
            static_inp (str, optional): _description_. Defaults to "SPG".
            l_inp_multi (list, optional): _description_. Defaults to [1,1]. First Value: Multispan in [m] and Second value:
        """
        self.statical_system = static_inp
        self.input_discret = num_input
        self.E = E_inp
        self.I = I_inp
        self.A = A_inp
        self.l = l_inp
        self.rho = rho_inp

        self.l_2 = l_inp_multi[0]

        self.absorber_node = []
        self.absorber_mass = []
        self.absorber_stiffness = []

        self.num_integral = _num_integral  # Check for the required accuracy:
        self.f_s = 1 / 10000

        # Length vector for numerical integration

        self.length = [
            self.l / self.num_integral * i for i in range(self.num_integral + 1)
        ]

        self.length_2 = [
            self.l_2 / self.num_integral * i for i in range(self.num_integral + 1)
        ]

    def calculation_params_spg(self):
        self.xi = [
            (i + 1) / (self.input_discret + 1) for i in range(self.input_discret)
        ]
        self.m_moment = np.zeros((self.num_integral, self.input_discret))

        for col in range(self.input_discret):
            for row in range(self.num_integral):
                A_Bearing = 1 - self.xi[col]
                B_Bearing = self.xi[col]
                if row <= self.xi[col] * (self.num_integral):
                    self.m_moment[row][col] = (
                        A_Bearing * row / self.num_integral * self.l
                    )

                else:
                    self.m_moment[row][col] = (
                        A_Bearing * row / self.num_integral * self.l
                        - (row / self.num_integral - self.xi[col]) * self.l
                    )

    def calculation_params_ctb(self):
        self.xi = [(i + 1) / self.input_discret for i in range(self.input_discret)]
        self.m_moment = np.zeros((self.num_integral, self.input_discret))

        for col in range(self.input_discret):
            for row in range(self.num_integral):
                A_Bearing = 1
                M_Bearing = self.xi[col] * self.l * A_Bearing
                if row <= int(self.xi[col] * self.num_integral):
                    self.m_moment[row][col] = (
                        M_Bearing - A_Bearing * row / self.num_integral * self.l
                    )
                else:
                    self.m_moment[row][col] = 0
        self.length = [
            self.l / self.num_integral * i for i in range(self.num_integral + 1)
        ]

    def calculation_params_two_span_girder(self):
        ## Calculating the M0-Moment-Curve for discrete single loads at each point
        self.xi = [
            (i + 1) / (self.input_discret + 1) for i in range(self.input_discret)
        ]
        self.m_moment_0 = np.zeros((self.num_integral + 1, self.input_discret))
        self.m_moment_1 = np.zeros(
            self.num_integral + 1
        )  # valid, if the bearing of the two-span girder is in the middle of the bridge (Usually fullfilled)
        self.m_moment = np.zeros((self.num_integral + 1, self.input_discret))

        for col in range(self.input_discret):
            for row in range(self.num_integral + 1):
                A_Bearing = 1 - self.xi[col]
                B_Bearing = self.xi[col]
                if row <= self.xi[col] * self.num_integral:
                    self.m_moment_0[row][col] = (
                        A_Bearing * row / self.num_integral * self.l
                    )
                else:
                    self.m_moment_0[row][col] = (
                        A_Bearing * row / self.num_integral * self.l
                        - (row / self.num_integral - self.xi[col]) * self.l
                    )
                if col == 0:
                    if row <= (self.num_integral + 1) / 2:
                        self.m_moment_1[row] = -row / self.num_integral * self.l * 0.5
                    else:
                        self.m_moment_1[row] = (
                            -self.l / 4
                            + (row - (self.num_integral) / 2)
                            / self.num_integral
                            * self.l
                            * 0.5
                        )

        self.m_moment_1[-1] = 0
        self.length_0 = [
            self.l / self.num_integral * i for i in range(self.num_integral)
        ]

        ## Calculation of the delta-coefficients
        integral_11 = [
            self.m_moment_1[i] * self.m_moment_1[i]
            for i in range(0, len(self.m_moment_1), 1)
        ]

        delta_11 = np.trapz(integral_11, self.length) * 1 / (self.E * self.I)

        x1 = np.zeros(self.input_discret)  # Statically indetermined
        self.delta_10_mat = np.zeros(self.input_discret)  # For testing purposes

        for col in range(self.input_discret):
            m_moment_inte = np.zeros(self.num_integral + 1)
            for i in range(self.num_integral + 1):
                m_moment_inte[i] = self.m_moment_0[i][col]

            integrant_10 = m_moment_inte * self.m_moment_1
            delta_10 = np.trapz(integrant_10, self.length) * 1 / (self.E * self.I)

            self.delta_10_mat[col] = delta_10
            x1[col] = -delta_10 / delta_11

            for i in range(self.num_integral + 1):
                self.m_moment[i][col] = (
                    self.m_moment_0[i][col] + x1[col] * self.m_moment_1[i]
                )

    def calculation_params_two_span_girder_SPG(self):
        ## Calculating the M0-Moment-Curve for discrete single loads at each point
        self.xi = [
            (i + 1) / (self.input_discret + 1) for i in range(self.input_discret)
        ]
        self.m_moment_0_1 = np.zeros(
            (self.num_integral + 1, self.input_discret)
        )  # First statically determined part system
        self.m_moment_0_2 = np.zeros(
            (self.num_integral + 1, self.input_discret)
        )  # Second statically determined part system

        self.m_moment_1_1 = np.zeros(self.num_integral + 1)
        self.m_moment_1_2 = np.zeros(self.num_integral + 1)

        self.m_moment = np.zeros(((self.num_integral + 1) * 2, self.input_discret * 2))

        # Calulating the moment-curves for the first span

        for col in range(self.input_discret):
            for row in range(self.num_integral + 1):
                A_Bearing = 1 - self.xi[col]
                B_Bearing = self.xi[col]
                if row <= self.xi[col] * self.num_integral:
                    self.m_moment_0_1[row][col] = (
                        A_Bearing * row / self.num_integral * self.l
                    )
                else:
                    self.m_moment_0_1[row][col] = (
                        A_Bearing * row / self.num_integral * self.l
                        - (row / self.num_integral - self.xi[col]) * self.l
                    )

                if col == 0:
                    self.m_moment_1_1[row] = -row / self.num_integral * self.l * 1 / self.l


                # Calculating the moment curves for the second span

        # Calculating the moment-curves for the second span
        for col in range(self.input_discret):
            for row in range(self.num_integral + 1):
                A_Bearing = 1 - self.xi[col]
                B_Bearing = self.xi[col]
                if row <= self.xi[col] * self.num_integral:
                    self.m_moment_0_2[row][col] = (
                        A_Bearing * row / self.num_integral * self.l_2
                    )
                else:
                    self.m_moment_0_2[row][col] = (
                        A_Bearing * row / self.num_integral * self.l_2
                        - (row / self.num_integral - self.xi[col]) * self.l_2
                    )

                if col == 0:
                    self.m_moment_1_2[row] = (
                            -1 + row / self.num_integral * self.l_2 * 1 / self.l_2
                        )

                        
        # Calculating the influence number for delta_11

        integral_11_1 = [
            self.m_moment_1_1[i] * self.m_moment_1_1[i]
            for i in range(0, len(self.m_moment_1_1), 1)
        ]

        delta_11_1 = np.trapz(integral_11_1, self.length) * 1 / (self.E * self.I)

        integral_11_2 = [
            self.m_moment_1_2[i] * self.m_moment_1_2[i]
            for i in range(0, len(self.m_moment_1_2), 1)
        ]

        delta_11_2 = np.trapz(integral_11_2, self.length_2) * 1 / (self.E * self.I)


        # Calulating the influence number for delta_10 for the first span

        x1 = np.zeros(self.input_discret)  # Statically indetermined moment an the middle bearing

        for col in range(self.input_discret):
            m_moment_inte_1 = np.zeros(self.num_integral + 1)
            m_moment_inte_2 = np.zeros(self.num_integral + 1)
            for i in range(self.num_integral + 1):
                m_moment_inte_1[i] = self.m_moment_0_1[i][col]
                m_moment_inte_2[i] = self.m_moment_0_2[i][col]

            integrant_10_1 = m_moment_inte_1 * self.m_moment_1_1
            delta_10_1 = np.trapz(integrant_10_1, self.length) * 1 / (self.E * self.I)

            # print("Statically determined moment")
            # print(self.m_moment_0_1)
            # print("Virtual Moment - 1")
            # print(self.m_moment_1_1)
            # print("Virtual Moment - 2")
            # print(self.m_moment_1_2)

            # print("delta 10 ")

            # print(delta_10_1)

            # print("delta 11")

            # print(delta_11_1 + delta_11_2)

            x1[col] = -(delta_10_1 / (delta_11_1 + delta_11_2))

            # print(x1)

            # Integration result for the first span

            for i in range(self.num_integral + 1):
                self.m_moment[i][col] = (
                    self.m_moment_0_1[i][col] + x1[col] * self.m_moment_1_1[i]
                )

            # Integration result for the second span

            for i in range(self.num_integral + 1, (self.num_integral + 1) * 2):
                self.m_moment[i][col] = (
                    0 + x1[col] * self.m_moment_1_2[int(i-(self.num_integral + 1))]
                )

        # Calculating the influence number for delta_10 for the second span
        x1 = np.zeros(self.input_discret)  # Statically indetermined moment an the middle bearing

        for col in range(self.input_discret,self.input_discret*2):
            m_moment_inte_1 = np.zeros(self.num_integral + 1)
            m_moment_inte_2 = np.zeros(self.num_integral + 1)
            for i in range(self.num_integral + 1):
                m_moment_inte_2[i] = self.m_moment_0_2[i][int(col-self.input_discret)]

            integrant_10_2 = m_moment_inte_2 * self.m_moment_1_2
            delta_10_2 = np.trapz(integrant_10_2, self.length_2) * 1 / (self.E * self.I)

            # print("Statically determined moment")
            # print(self.m_moment_0_2)
            # print("Virtual Moment - 1")
            # print(self.m_moment_1_1)
            # print("Virtual Moment - 2")
            # print(self.m_moment_1_2)

            # print("delta 10 ")

            # print(delta_10_2)

            # print("delta 11")

            # print(delta_11_1 + delta_11_2)

            x1[int(col-self.input_discret)] = -(delta_10_2 / (delta_11_1 + delta_11_2))

            # print(x1)

            # Integration result for the first span

            for i in range(self.num_integral + 1):
                self.m_moment[i][col] = (
                    x1[int(col-self.input_discret)] * self.m_moment_1_1[i]
                )

            # Integration result for the second span

            for i in range(self.num_integral + 1, (self.num_integral+1) * 2):
                self.m_moment[i][col] = (  
                    self.m_moment_0_2[i-(self.num_integral + 1)][int(col-self.input_discret)] + x1[int(col-self.input_discret)] * self.m_moment_1_2[int(i-(self.num_integral + 1))]
                )

    def mass_matrix(self):
        if (self.statical_system == "TSG_H"):
            self.M_matrix = np.zeros(
                (
                    self.input_discret*2 + len(self.absorber_node),
                    self.input_discret*2 + len(self.absorber_node),
                )
            )

            m_i_1 = self.rho * self.l * self.A / (self.input_discret + 1)
            m_i_2 = self.rho * self.l_2 * self.A / (self.input_discret + 1)
            
            for row in range(self.input_discret):
                self.M_matrix[row][row] = m_i_1
                self.M_matrix[row+self.input_discret][row+self.input_discret] = m_i_2

            for row in range(len(self.absorber_node)):
                self.M_matrix[row + self.input_discret*2][
                    row + self.input_discret * 2
                ] = self.absorber_mass[row]

        else:
            self.M_matrix = np.zeros(
                (
                    self.input_discret + len(self.absorber_node),
                    self.input_discret + len(self.absorber_node),
                )
            )
            m_i = self.rho * self.l * self.A / (self.input_discret + 1)

            for row in range(self.input_discret):
                self.M_matrix[row][row] = m_i

            for row in range(len(self.absorber_node)):
                self.M_matrix[row + self.input_discret][
                    row + self.input_discret
                ] = self.absorber_mass[row]

    def single_span_girder(self):
        self.mass_matrix()
        EI = self.E * self.I
        self.calculation_params_spg()
        integral = np.zeros(self.num_integral+1)
        self.delta_i = np.zeros((self.input_discret, self.input_discret))

        for row in range(self.input_discret):
            for col in range(self.input_discret):
                for i in range(self.m_moment.shape[0]):
                    integral[i] = self.m_moment[i][row] * self.m_moment[i][col]
                self.delta_i[row][col] = (1 / EI) * np.trapz(integral, self.length)

        self.K_matrix = np.linalg.inv(self.delta_i)

    def cantilever_beam(self):
        self.mass_matrix()
        EI = self.E * self.I
        self.calculation_params_ctb()
        integral = np.zeros(self.num_integral+1)
        self.delta_i = np.zeros((self.input_discret, self.input_discret))

        for row in range(self.input_discret):
            for col in range(self.input_discret):
                for i in range(self.m_moment.shape[0]):
                    integral[i] = self.m_moment[i][row] * self.m_moment[i][col]
                self.delta_i[row][col] = (1 / EI) * np.trapz(integral, self.length)

        self.K_matrix = np.linalg.inv(self.delta_i)

    def two_span_girder(self):
        self.mass_matrix()
        EI = self.E * self.I
        self.calculation_params_two_span_girder()

        integral = np.zeros(self.num_integral + 1)
        self.delta_i = np.zeros((self.input_discret, self.input_discret))

        for row in range(self.input_discret):
            for col in range(self.input_discret):
                for i in range(self.m_moment.shape[0]):
                    integral[i] = (
                        self.m_moment[i][row] * self.m_moment_0[i][col]
                    )  # m_moment_0 ->
                    # Virtual moment at the statically determined system
                    # Virtual moment is equal to the moment_0
                self.delta_i[row][col] = (1 / EI) * np.trapz(integral, self.length)

        # Define your threshold for deflection values below which rows and columns are removed

        threshold = 1e-15

        # Calculate the absolute values of the matrices
        abs_GSM = np.abs(self.delta_i)
        abs_GMM = np.abs(self.M_matrix)

        # Create masks to identify rows and columns with values above the threshold
        rows_to_keep_GSM = np.any(abs_GSM >= threshold, axis=1)
        cols_to_keep_GSM = np.any(abs_GSM >= threshold, axis=0)

        # Perform row reduction based on the threshold for the global stiffness matrix
        rrgsm = self.delta_i[rows_to_keep_GSM]
        crgsm = rrgsm[:, cols_to_keep_GSM]
        self.delta_i = crgsm

        # Perform row reduction based on the threshold for the global mass matrix
        rgmm = self.M_matrix[rows_to_keep_GSM]
        crgmm = rgmm[:, cols_to_keep_GSM]
        self.M_matrix = crgmm

        self.K_matrix = np.linalg.inv(self.delta_i)
   

    def two_span_girder_hinged(self):
        print("Two span girder with a hinge")
        self.mass_matrix()
        EI = self.E * self.I
        self.calculation_params_two_span_girder_SPG()

        integral_1 = np.zeros(self.num_integral + 1)
        integral_2 = np.zeros(self.num_integral + 1)

        self.delta_i = np.zeros((self.input_discret * 2, self.input_discret * 2))

        for row in range(self.input_discret*2):             # Corresponds to the loading node (Single load 1)
            for col in range(self.input_discret*2):         # Corresponds to the reaction node (Node, where the deflection is calculated (Virtual Moment!))
                for i in range(int(self.m_moment.shape[0]/2)):
                    if (col < self.input_discret):
                        # Case 1: The deformations are calculated for the first n = self.input_discrete masses in the span 1
                        # Therefore the integral in the second span is zero, because the moment curves are decoupled due to the hinge 
                        # in the midspan
                        integral_1[i] = (
                            self.m_moment[i][row] * self.m_moment_0_1[i][col-self.input_discret]
                        )  
                        integral_2[i] = (
                            self.m_moment[i+int(self.m_moment.shape[0]/2)][row] * 0
                        )
                    else:
                        # Case 2: The deformations are calculated for the second n = self.input_discrete masses in the span 2
                        # Therefore the integral in the first span is zero, because the moment curves are decoupled due to the hinge 
                        # in the midspan. 
                        integral_1[i] = (
                            self.m_moment[i][row] * 0
                        )  
                        integral_2[i] = (
                            self.m_moment[i+int(self.m_moment.shape[0]/2)][row] *self.m_moment_0_2[i][col-self.input_discret]
                        )

                #     print(self.m_moment[i+int(self.m_moment.shape[0]/2)][row])

                self.delta_i[row][col] = (1 / EI) * (np.trapz(integral_1, self.length) + np.trapz(integral_2, self.length_2) ) 



        self.K_matrix = np.linalg.inv(self.delta_i)

    def compute_eigenfrequencies(self):
        try:
            eigenvalues, eigenvectors = eigh(self.K_matrix, self.M_matrix, subset_by_index=(0, 20 - 1))
        except:
            eigenvalues, eigenvectors = eigh(self.K_matrix, self.M_matrix)      # Eigenvalues with Number of DOFS below 20

        num_eigenval = len(eigenvalues)

        self.eigenfrequencies = np.sqrt(np.abs(eigenvalues))

        normalized_modes = eigenvectors / np.abs(eigenvectors).max(axis=0)

        self.eigenmodes_matrix_storage = normalized_modes

        self.eigenmodes_matrix = normalized_modes


        num_rows, num_columns = self.eigenmodes_matrix.shape

        # Define the number of zeros to add at the beginning and end of each column
        num_zeros = 1  # You want to add one zero at each end

        # Create a new matrix with the desired structure
        new_num_rows = num_rows + 2 * num_zeros  # Calculate the new number of rows
        new_matrix = np.zeros((new_num_rows, num_columns))  # Initialize a new matrix with zeros

        # Copy the original matrix into the center of the new matrix
        new_matrix[num_zeros:num_zeros + num_rows, :] = self.eigenmodes_matrix

        self.eigenmodes_matrix = new_matrix

        if self.statical_system == "TSG_H":     # Adding a zero deflection point, because of the vertical bearing

            new_num_rows = self.eigenmodes_matrix.shape[0] + 1  # Calculate the new number of rows
            new_matrix = np.zeros((new_num_rows, num_columns))  # Initialize a new matrix with zeros

            # Copy the first half of the original matrix between the first row and the middle row
            for row in range(0, 1 + self.input_discret):
                for col in range(num_eigenval):
                    new_matrix[row][col] = self.eigenmodes_matrix[row][col]
            try:
                for row in range(1 + self.input_discret, self.input_discret*2+3):
                    print(row)
                    for col in range(num_eigenval):
                        new_matrix[row+1][col] =  self.eigenmodes_matrix[row][col]
            except:
                pass

            for row in range(1 + self.input_discret, self.input_discret*2+2):
                print(row)
                for col in range(num_eigenval):
                    new_matrix[row+1][col] =  self.eigenmodes_matrix[row][col]


            self.eigenmodes_matrix = new_matrix

            # Length-Vector for the whole structure

            self.len_plotting = np.zeros(2*self.input_discret+3)

            for i in range(0,self.input_discret+1,1):
                self.len_plotting[i] = self.l / (self.input_discret+1) * i

            self.len_plotting[self.input_discret+1] = self.l

            for i in range(0,self.input_discret+1,1):
                self.len_plotting[i+self.input_discret+2] = self.l + self.l_2 / (self.input_discret+1) * (i+1)


        else:
            # Length-Vector for the whole structure
            self.len_plotting = np.zeros(self.input_discret+2)

            for i in range(0, self.input_discret+1, 1):
                self.len_plotting[i] = self.l / (self.input_discret+1) * i

            self.len_plotting[self.input_discret+1] = self.l


        return self.eigenfrequencies, self.eigenmodes_matrix


    def modal_matrices(self):
        self.M_trans = np.dot(np.transpose(self.eigenmodes_matrix_storage),np.dot(self.M_matrix,self.eigenmodes_matrix_storage))
        self.K_trans = np.dot(np.transpose(self.eigenmodes_matrix_storage),np.dot(self.K_matrix,self.eigenmodes_matrix_storage))


    def print_modes_matrix(self):
        print("Circular eigenfrequencies")
        print(self.eigenfrequencies)

        print("Eigenfrequencies")

        print(self.eigenfrequencies / (2 * np.pi))

        print("Normalized eigenmodes")

        for row in self.eigenmodes_matrix:
            formatted_row = ["{:.4f}".format(item) for item in row]
            print(" | ".join(formatted_row))
            print("-" * (len(formatted_row) * 8))  # Print dashes for clarity



import numpy as np
from scipy.optimize import newton
import sqlite3
from scipy import integrate
from scipy.misc import derivative


class eigenmodes:
    def __init__(self):
        self.num_x = 10000

        self.lambda_cantilever = np.zeros(20)
        self.a_lambda_cantilever = np.zeros(20)
        self.j1_cantilever = np.zeros(20)
        self.y_cantilever = np.zeros((20, self.num_x))
        self.y_xx_cantilever = np.zeros((20, self.num_x))

        self.lambda_clamped_hinged = np.zeros(20)
        self.a_lambda_clamped_hinged = np.zeros(20)
        self.j1_clamped_hinged = np.zeros(20)
        self.y_clamped_hinged = np.zeros((20, self.num_x))
        self.y_xx_clamped_hinged = np.zeros((20, self.num_x))

        self.lambda_hinged_hinged = np.zeros(20)
        self.j1_hinged_hinged = np.zeros(20)
        self.y_hinged_hinged = np.zeros((20, self.num_x))
        self.y_xx_hinged_hinged = np.zeros((20, self.num_x))

        self.lambda_clamped_clamped = np.zeros(20)
        self.a_lambda_clamped_clamped = np.zeros(20)
        self.j1_clamped_clamped = np.zeros(20)
        self.y_clamped_clamped = np.zeros((20, self.num_x))
        self.y_xx_clamped_clamped = np.zeros((20, self.num_x))

        self.J_2_hinged_hinged = np.zeros(20)
        self.J_2_clamped_hinged = np.zeros(20)
        self.J_2_clamped_clamped = np.zeros(20)
        self.J_2_cantilever = np.zeros(20)

        self.J_3_hinged_hinged = np.zeros(20)
        self.J_3_clamped_hinged = np.zeros(20)
        self.J_3_clamped_clamped = np.zeros(20)
        self.J_3_cantilever = np.zeros(20)

    # Functions for the lambda-values

    def equation_cantilever_cal(self, lambda_val):
        return np.cosh(lambda_val) * np.cos(lambda_val) + 1

    def equation_clamped_hinged_cal(self, lambda_val):
        return np.tanh(lambda_val) - np.tan(lambda_val)

    def equation_hinged_hinged_cal(self, lambda_val):
        return np.sin(lambda_val)

    def equation_clamped_clamped_cal(self, lambda_val):
        return np.cosh(lambda_val) * np.cos(lambda_val) - 1

    # Functions for a lambda

    def a_lambda_cantilever_cal(self, lambda_values):
        return (np.sinh(lambda_values) + np.sin(lambda_values)) * (
            np.cosh(lambda_values) + np.cos(lambda_values)
        ) ** (-1)

    def a_lambda_clamped_hinged_cal(self, lambda_values):
        return (np.sinh(lambda_values) - np.sin(lambda_values)) * (
            np.cosh(lambda_values) - np.cos(lambda_values)
        ) ** (-1)

    def a_lambda_hinged_hinged_cal(self, lambda_values):
        return (lambda_values) * 0

    def a_lambda_clamped_clamped_cal(self, lambda_values):
        return (np.sinh(lambda_values) - np.sin(lambda_values)) * (
            np.cosh(lambda_values) - np.cos(lambda_values)
        ) ** (-1)

    def solving_eigenvalues(self):
        guess_cantilever = 0
        guess_clamped_hinged = 0
        guess_hinged_hinged = 0
        guess_clamped_clamped = 0

        for i in range(0, 20, 1):
            if i == 0:
                guess_cantilever = 1.874
                guess_clamped_hinged = 3.92
                guess_clamped_clamped = 4.73

            elif i == 1:
                guess_cantilever = 4.68
                guess_clamped_hinged = 7.06
                guess_clamped_clamped = 7.85
            elif i == 2:
                guess_cantilever = 7.84
                guess_clamped_hinged = 10.21
                guess_clamped_clamped = 10.99
            else:
                guess_cantilever = (
                    2 * (i + 1) - 1
                ) * np.pi / 2 - 0.02  # Important: i+1, because of the formulation for the eigenfrequencies
                guess_clamped_hinged = (
                    i + 1 + 0.25
                ) * np.pi  # Otherwise the convergance for the eigenfrequencies won't work, because a lower frequency value
                guess_clamped_clamped = (i + 1 + 0.5) * np.pi  # is calculated

            guess_hinged_hinged = (
                i + 1
            ) * np.pi  # The hinged-hinged always equals to i * np.pi

            print(guess_hinged_hinged)

            self.lambda_cantilever[i] = newton(
                self.equation_cantilever_cal, guess_cantilever
            )
            self.lambda_clamped_hinged[i] = newton(
                self.equation_clamped_hinged_cal, guess_clamped_hinged
            )
            self.lambda_hinged_hinged[i] = newton(
                self.equation_hinged_hinged_cal, guess_hinged_hinged
            )
            self.lambda_clamped_clamped[i] = newton(
                self.equation_clamped_clamped_cal, guess_clamped_clamped
            )

    def calculation_a_lambda(self):
        for i in range(0, 20, 1):
            self.a_lambda_cantilever[i] = self.a_lambda_cantilever_cal(
                self.lambda_cantilever[i]
            )
            self.a_lambda_clamped_hinged[i] = self.a_lambda_clamped_hinged_cal(
                self.lambda_clamped_hinged[i]
            )
            self.a_lambda_clamped_clamped[i] = self.a_lambda_clamped_clamped_cal(
                self.lambda_clamped_clamped[i]
            )

    def calculating_j1(self):
        self.x_values = np.linspace(0, 1, num=self.num_x)
        for i in range(0, 20, 1):
            for j in range(0, len(self.x_values), 1):
                self.y_hinged_hinged[i][j] = np.sin(
                    self.lambda_hinged_hinged[i] * self.x_values[j]
                )

                self.y_clamped_clamped[i][j] = (
                    np.sin(self.lambda_clamped_clamped[i] * self.x_values[j])
                    - np.sinh(self.lambda_clamped_clamped[i] * self.x_values[j])
                    + self.a_lambda_clamped_clamped[i]
                    * (
                        np.cosh(self.lambda_clamped_clamped[i] * self.x_values[j])
                        - np.cos(self.lambda_clamped_clamped[i] * self.x_values[j])
                    )
                )
                self.y_clamped_hinged[i][j] = (
                    np.sin(self.lambda_clamped_hinged[i] * self.x_values[j])
                    - np.sinh(self.lambda_clamped_hinged[i] * self.x_values[j])
                    + self.a_lambda_clamped_hinged[i]
                    * (
                        np.cosh(self.lambda_clamped_hinged[i] * self.x_values[j])
                        - np.cos(self.lambda_clamped_hinged[i] * self.x_values[j])
                    )
                )

                self.y_cantilever[i][j] = (
                    np.sin(self.lambda_cantilever[i] * self.x_values[j])
                    - np.sinh(self.lambda_cantilever[i] * self.x_values[j])
                    + self.a_lambda_cantilever[i]
                    * (
                        np.cosh(self.lambda_cantilever[i] * self.x_values[j])
                        - np.cos(self.lambda_cantilever[i] * self.x_values[j])
                    )
                )

                self.j1_hinged_hinged[i] = integrate.simps(
                    self.y_hinged_hinged[i][:] ** 2, x=self.x_values
                )
                self.j1_clamped_hinged[i] = integrate.simps(
                    self.y_clamped_hinged[i][:] ** 2, x=self.x_values
                )
                self.j1_clamped_clamped[i] = integrate.simps(
                    self.y_clamped_clamped[i][:] ** 2, x=self.x_values
                )
                self.j1_cantilever[i] = integrate.simps(
                    self.y_cantilever[i][:] ** 2, x=self.x_values
                )

    def numerical_derivative(self, data, delta_x):
        """
        Compute the numerical derivative of the given data using central difference quotient.
        :param data: List of function values.
        :param delta_x: Spacing between the x-values.
        :return: List of derivative values.
        """
        derivative = []
        for i in range(1, len(data) - 1):
            df = (data[i + 1] - data[i - 1]) / (2 * delta_x)
            derivative.append(df)
        return derivative

    def second_derivative(self, data, delta_x):
        # Compute the first derivative
        first_derivative = self.numerical_derivative(data, delta_x)

        # Compute the second derivative from the first derivative
        return self.numerical_derivative(first_derivative, delta_x)

    def calculating_j2(self):
        for i in range(0, 20, 1):
            y_xx_temp_cantilever = self.second_derivative(
                self.y_cantilever[i][:], 1 / self.num_x
            )
            self.y_xx_cantilever[i][2:-2] = y_xx_temp_cantilever

            y_xx_temp_hinged_hinged = self.second_derivative(
                self.y_hinged_hinged[i][:], 1 / self.num_x
            )
            self.y_xx_hinged_hinged[i][2:-2] = y_xx_temp_hinged_hinged

            y_xx_temp_clamped_hinged = self.second_derivative(
                self.y_clamped_hinged[i][:], 1 / self.num_x
            )
            self.y_xx_clamped_hinged[i][2:-2] = y_xx_temp_clamped_hinged

            y_xx_temp_clamped_clamped = self.second_derivative(
                self.y_clamped_clamped[i][:], 1 / self.num_x
            )
            self.y_xx_clamped_clamped[i][2:-2] = y_xx_temp_clamped_clamped

            self.J_2_cantilever[i] = integrate.simps(
                self.y_xx_cantilever[i][:] ** 2, x=self.x_values
            )
            self.J_2_hinged_hinged[i] = integrate.simps(
                self.y_xx_hinged_hinged[i][:] ** 2, x=self.x_values
            )
            self.J_2_clamped_hinged[i] = integrate.simps(
                self.y_xx_clamped_hinged[i][:] ** 2, x=self.x_values
            )

            self.J_2_clamped_clamped[i] = integrate.simps(
                self.y_xx_clamped_clamped[i][:] ** 2, x=self.x_values
            )

    def calculating_j3(self):
        for i in range(0, 20, 1):
            self.J_3_cantilever[i] = integrate.simps(
                    self.y_cantilever[i][:], x=self.x_values
                )
            self.J_3_clamped_clamped[i] = integrate.simps(
                    self.y_clamped_clamped[i][:], x=self.x_values
                )
            self.J_3_clamped_hinged[i] = integrate.simps(
                    self.y_clamped_hinged[i][:], x=self.x_values
                )
            self.J_3_hinged_hinged[i] =integrate.simps(
                    self.y_hinged_hinged[i][:], x=self.x_values
                )

    def save_to_database(self, db_name="Database/eigenmodes.db"):
        # Connect to the SQLite database
        conn = sqlite3.connect(db_name)
        cursor = conn.cursor()

        # Create tables
        cursor.execute(
            """
            CREATE TABLE IF NOT EXISTS eigenvalues (
                id INTEGER PRIMARY KEY,
                lambda_cantilever REAL,
                lambda_clamped_hinged REAL,
                lambda_hinged_hinged REAL,
                lambda_clamped_clamped REAL
            )
            """
        )

        cursor.execute(
            """
            CREATE TABLE IF NOT EXISTS a_lambda (
                id INTEGER PRIMARY KEY,
                a_lambda_cantilever REAL,
                a_lambda_clamped_hinged REAL,
                a_lambda_clamped_clamped REAL
            )
            """
        )

        cursor.execute(
            """
            CREATE TABLE IF NOT EXISTS J1 (
                id INTEGER PRIMARY KEY,
                J1_cantilever REAL,
                J1_clamped_hinged REAL,
                J1_hinged_hinged REAL,
                J1_clamped_clamped REAL
            )
            """
        )

        cursor.execute(
            """
            CREATE TABLE IF NOT EXISTS J2 (
                id INTEGER PRIMARY KEY,
                J2_cantilever REAL,
                J2_clamped_hinged REAL,
                J2_hinged_hinged REAL,
                J2_clamped_clamped REAL
            )
            """
        )

        cursor.execute(
            """
            CREATE TABLE IF NOT EXISTS J3 (
                id INTEGER PRIMARY KEY,
                J3_cantilever REAL,
                J3_clamped_hinged REAL,
                J3_hinged_hinged REAL,
                J3_clamped_clamped REAL
            )
            """
        )

        # Delete existing data to avoid duplicate entries
        cursor.execute("DELETE FROM eigenvalues")
        cursor.execute("DELETE FROM a_lambda")
        cursor.execute("DELETE FROM J1")
        cursor.execute("DELETE FROM J2")
        cursor.execute("DELETE FROM J3")

        # Insert data
        for i in range(20):
            cursor.execute(
                """
                INSERT INTO eigenvalues (lambda_cantilever, lambda_clamped_hinged, lambda_hinged_hinged, lambda_clamped_clamped)
                VALUES (?, ?, ?, ?)
                """,
                (
                    self.lambda_cantilever[i],
                    self.lambda_clamped_hinged[i],
                    self.lambda_hinged_hinged[i],
                    self.lambda_clamped_clamped[i],
                ),
            )

            cursor.execute(
                """
                INSERT INTO a_lambda (a_lambda_cantilever, a_lambda_clamped_hinged, a_lambda_clamped_clamped)
                VALUES (?, ?,?)
                """,
                (self.a_lambda_cantilever[i], self.a_lambda_clamped_hinged[i],self.a_lambda_clamped_clamped[i])
            )

            cursor.execute(
                """
                INSERT INTO J1 (J1_cantilever, J1_clamped_hinged, J1_hinged_hinged, J1_clamped_clamped)
                VALUES (?, ?, ?, ?)
                """,
                (
                    self.j1_cantilever[i],
                    self.j1_clamped_hinged[i],
                    self.j1_hinged_hinged[i],
                    self.j1_clamped_clamped[i],
                ),
            )

            cursor.execute(
                """
                INSERT INTO J2 (J2_cantilever, J2_clamped_hinged, J2_hinged_hinged, J2_clamped_clamped)
                VALUES (?, ?, ?, ?)
                """,
                (
                    self.J_2_cantilever[i],
                    self.J_2_clamped_hinged[i],
                    self.J_2_hinged_hinged[i],
                    self.J_2_clamped_clamped[i],
                ),
            )

            cursor.execute(
                """
                INSERT INTO J3 (J3_cantilever, J3_clamped_hinged, J3_hinged_hinged, J3_clamped_clamped)
                VALUES (?, ?, ?, ?)
                """,
                (
                    self.J_3_cantilever[i],
                    self.J_3_clamped_hinged[i],
                    self.J_3_hinged_hinged[i],
                    self.J_3_clamped_clamped[i],
                ),
            )

        # Commit changes and close connection
        conn.commit()
        conn.close()


test_class = eigenmodes()

test_class.solving_eigenvalues()

print("Testing for the cantilever")
print(test_class.lambda_cantilever)
print("Testing for the clamped hinged beam")
print(test_class.lambda_clamped_hinged)

print("Testing for the hinged hinged beam")
print(test_class.lambda_hinged_hinged)
print("Testing for the clamped clamped beam")
print(test_class.lambda_clamped_clamped)

test_class.calculation_a_lambda()
test_class.calculating_j1()
test_class.calculating_j2()
test_class.calculating_j3()

test_class.save_to_database()



# DEPENDENCIES
import copy  # Allows us to create copies of objects in memory
import math  # Math functionality
import numpy as np  # Numpy for working with arrays
import matplotlib.pyplot as plt  # Plotting functionality
import matplotlib.colors  # For colormap functionality
import ipywidgets as widgets
from glob import glob  # Allows check that file exists before import
from numpy import genfromtxt  # For importing structure data from csv
import pandas as pd

try:
    from InputData import Input
    from Materials import Material
    from Elements import Rope_Elements_III
    from Elements import BarElements_I
except:
    from KIB_LAP.FACHWERKEBEN.InputData import Input
    from KIB_LAP.FACHWERKEBEN.Materials import Material
    from KIB_LAP.FACHWERKEBEN.Elements import Rope_Elements_III
    from KIB_LAP.FACHWERKEBEN.Elements import BarElements_I

class IterationClass:
    def __init__(self, use_iteration=False):
        print("INIT")
        ##________ Subclasses __________##
        self.Inp = Input()
        self.Mat = Material()
        self.CableNonlinear = Rope_Elements_III(self.Inp)
        self.BarLinear = BarElements_I(self.Inp)


        self.swt = False

        ##___ Iteration parameters _________##

        self.nForceIncrements = 10
        self.convThreshold = 1  # (N) Threshold on average percentage increase in incremental deflection

        self.checkSlackAfter = 80

        # Member Types

        self.memberType = []
        self.ClassifyMemberType()

        ##____ Containers_____##
        # Initialise a container to hold the set of global displacements for each external load increment
        self.UG_FINAL = np.empty([self.Inp.nDoF, 0])

        # Initialise a container to hold the set of internal forces for each external load increment
        self.FI_FINAL = np.empty([self.Inp.nDoF, 0])

        # Initialise a container to hold the set of axial forces for each external load increment
        self.EXTFORCES = np.empty([self.Inp.nDoF, 0])

        # Initialise a container to hold the set of axial forces for each external load increment
        self.MBRFORCES = np.empty([len(self.Inp.members), 0])

        # Initialise global disp vector
        self.UG = np.zeros(
            [self.Inp.nDoF, 1]
        )  # Initialise global displacement vector to zero (undeformed state)

        # Calculate initial transformation matrices for all members based on undeformed position
        self.TMs = self.calculateTransMatrices(self.UG)

        # Init point loads to global force vector

        self.AddPointLoadsGlobal()


        # Calculate initial lengths

        self.calculateInitialLengths()
        self.SelfweigthLoadVector()

        # Calculate internal force system based on any pre-tension in members
        self.F_pre = self.initPretension()

        # Initialise a container to store incremental displacements calculated for each iteration [Xa], [Xb] etc.
        self.UG_inc = np.empty([self.Inp.nDoF, 0])
        self.UG_inc = np.append(
            self.UG_inc, self.UG, axis=1
        )  # Add the initial (zero) displacement record

        # Initialise a container to store incremental internal forces calculated for each iteration [Fa], [Fb] etc.
        self.F_inc = np.empty([self.Inp.nDoF, 0])
        print(self.F_inc)
        print(self.F_pre)
        self.F_inc = np.append(
            self.F_inc, self.F_pre, axis=1
        )  # Add the initial pre-tension force record

        if use_iteration:
            self.MainConvergenceLoop()
        else:
            self.SolveLinear_NoIteration(treat_cables_as_bars=True)

    def ClassifyMemberType(self):
        print("Classify Member type")
        for n,m in enumerate(self.Inp.members):   
            #Initially assume all members are bars 
            self.memberType.append('b')

                
            #Check if member is a cable
            for c in self.Inp.cables:         
                if(m[0] in c and m[1] in c):
                    self.memberType[n] = 'c' 

    def AddPointLoadsGlobal(self):
        self.forceVector = np.zeros((len(self.Inp.nodes) * 2, 1))

        if len(self.Inp.forceLocationData) > 0:
            # Split force location data
            try:
                forcedNodes = self.Inp.forceLocationData[:, 0].astype(
                    int
                )  # Ensure these are integers)
                xForceIndizes = 2 * forcedNodes - 2
                yForceIndizes = 2 * forcedNodes - 1

                ForceP = self.Inp.forceLocationData[:, 1].reshape(-1, 1)
                print("FORCE P")
                print(ForceP)

                # Assign forces to degrees of freedom
                for i in range(0,len(ForceP),1):

                    if self.Inp.forceDirections[i]  == "x":
                        self.forceVector[xForceIndizes[i]] = ForceP[i]
                    elif self.Inp.forceDirections[i] == "y":
                        self.forceVector[yForceIndizes[i]] = ForceP[i]
            except:
                pass

    def calculateInitialLengths(self):
        self.lengths = np.zeros(len(self.Inp.members))
        for n, mbr in enumerate(self.Inp.members):

            # Calculate undeformed length of member
            node_i = mbr[0]  # Node number for node i of this member
            node_j = mbr[1]  # Node number for node j of this member
            ix = self.Inp.nodes[node_i - 1][0]  # x-coord for node i
            iy = self.Inp.nodes[node_i - 1][1]  # y-coord for node i
            jx = self.Inp.nodes[node_j - 1][0]  # x-coord for node j
            jy = self.Inp.nodes[node_j - 1][1]  # y-coord for node j

            dx = jx - ix  # x-component of vector along member
            dy = jy - iy  # y-component of vector along member
            length = math.sqrt(dx**2 + dy**2)  # Magnitude of vector (length of member)
            if (length == 0):
                print("Length = 0 at index ",  n )
            self.lengths[n] = length

    def initPretension(self):
        """
        P = axial pre-tension specified for each bar
        Calculate the force vector [F_pre] for each bar [F_pre] = [T'][AA'][P]
        Combine into an overal vector representing the internal force system and return
        """
        self.F_pre = np.array(
            [np.zeros(len(self.forceVector))]
        ).T  # Initialse internal force vector

        for n, mbr in enumerate(self.Inp.members):
            node_i = mbr[0]  # Node number for node i of this member
            node_j = mbr[1]  # Node number for node j of this member

            # Index of DoF for this member
            ia = 2 * node_i - 2  # horizontal DoF at node i of this member
            ib = 2 * node_i - 1  # vertical DoF at node i of this member
            ja = 2 * node_j - 2  # horizontal DoF at node j of this member
            jb = 2 * node_j - 1  # vertical DoF at node j of this member

            # Determine internal pre-tension in global coords
            TM = self.TMs[n, :, :]
            AAp = np.array([[1], [0]])
            P = self.Mat.P0[n]
            F_pre_global = np.matmul(TM.T, AAp) * P

            # Add member pre-tension to overall record
            self.F_pre[ia, 0] = self.F_pre[ia, 0] + F_pre_global[0][0]
            self.F_pre[ib, 0] = self.F_pre[ib, 0] + F_pre_global[1][0]
            self.F_pre[ja, 0] = self.F_pre[ja, 0] + F_pre_global[2][0]
            self.F_pre[jb, 0] = self.F_pre[jb, 0] + F_pre_global[3][0]

        return self.F_pre

    def calculateTransMatrices(self, UG):
        """
        Optimized:
        - Bars ('b'): transformation is constant (undeformed geometry) -> keep initial TM
        - Cables ('c'): update TM each iteration based on deformed geometry

        Returns array shape (nMembers, 2, 4)
        """

        nM = len(self.Inp.members)

        # -------------------------------------------------------
        # Create constant (initial) TMs once (for bars AND cables)
        # -------------------------------------------------------
        if not hasattr(self, "TMs_const") or self.TMs_const is None:
            self.TMs_const = np.zeros((nM, 2, 4), dtype=float)

            for n, mbr in enumerate(self.Inp.members):
                node_i = int(mbr[0])
                node_j = int(mbr[1])

                ix = float(self.Inp.nodes[node_i - 1, 0])
                iy = float(self.Inp.nodes[node_i - 1, 1])
                jx = float(self.Inp.nodes[node_j - 1, 0])
                jy = float(self.Inp.nodes[node_j - 1, 1])

                TM0 = self.CableNonlinear.calculateTransMatrix([ix, iy], [jx, jy])
                self.TMs_const[n, :, :] = TM0

        # start from constant TMs
        TMs = self.TMs_const.copy()

        # -------------------------------------------------------
        # Update ONLY cables
        # -------------------------------------------------------
        for n, mbr in enumerate(self.Inp.members):
            if self.memberType[n] != "c":
                continue  # bars: keep constant TM

            node_i = int(mbr[0])
            node_j = int(mbr[1])

            ia = 2 * node_i - 2
            ib = 2 * node_i - 1
            ja = 2 * node_j - 2
            jb = 2 * node_j - 1

            # deformed positions = initial + cumulative displacements
            ix = float(self.Inp.nodes[node_i - 1, 0]) + float(UG[ia, 0])
            iy = float(self.Inp.nodes[node_i - 1, 1]) + float(UG[ib, 0])
            jx = float(self.Inp.nodes[node_j - 1, 0]) + float(UG[ja, 0])
            jy = float(self.Inp.nodes[node_j - 1, 1]) + float(UG[jb, 0])

            TM = self.CableNonlinear.calculateTransMatrix([ix, iy], [jx, jy])
            TMs[n, :, :] = TM

        return TMs

    def buildStructureStiffnessMatrix(self, UG,TMs):
        """
        Standard construction of Primary and Structure stiffness matrix
        Construction of non-linear element stiffness matrix handled in a child function
        """
        Kp = np.zeros(
            [self.Inp.nDoF, self.Inp.nDoF]
        )  # Initialise the primary stiffness matrix

        # store spring stiffness diagonal (for equilibrium check)
        self.Kspring_diag = np.zeros(self.Inp.nDoF, dtype=float)


        for n, mbr in enumerate(self.Inp.members):
            node_i = mbr[0]  # Node number for node i of this member
            node_j = mbr[1]  # Node number for node j of this member

            # Construct (potentially) non-linear element stiffness matrix

            # [K11, K12, K21, K22] = self.CableNonlinear.buildElementStiffnessMatrix(
            #     n, UG, TMs, self.lengths, self.Mat.P0, self.Mat.E, self.Mat.A
            # )

            if self.memberType[n] == 'c':
                # cable / nonlinear
                [K11, K12, K21, K22] = self.CableNonlinear.buildElementStiffnessMatrix(
                    n, UG, TMs, self.lengths, self.Mat.P0, self.Mat.E, self.Mat.A
                )
            else:
                # bar / linear (Theorie I. Ordnung)
                [K11, K12, K21, K22] = self.BarLinear.buildElementStiffnessMatrix(
                    n, UG, None, None, self.Mat.P0, self.Mat.E, self.Mat.A
                )


            # Primary stiffness matrix indices associated with each node
            # i.e. node 1 occupies indices 0 and 1 (accessed in Python with [0:2])
            ia = 2 * node_i - 2  # index 0
            ib = 2 * node_i - 1  # index 1
            ja = 2 * node_j - 2  # index 2
            jb = 2 * node_j - 1  # index 3
            
            Kp[ia : ib + 1, ia : ib + 1] = Kp[ia : ib + 1, ia : ib + 1] + K11
            Kp[ia : ib + 1, ja : jb + 1] = Kp[ia : ib + 1, ja : jb + 1] + K12
            Kp[ja : jb + 1, ia : ib + 1] = Kp[ja : jb + 1, ia : ib + 1] + K21
            Kp[ja : jb + 1, ja : jb + 1] = Kp[ja : jb + 1, ja : jb + 1] + K22

        # Add springs

        if len(self.Inp.springLocationData) > 0:
            # Split force location data
            try:
                forcedNodes = self.Inp.springLocationData[:, 1].astype(
                    int
                )  # Ensure these are integers)
                xSpringIndizes = 2 * forcedNodes - 2
                ySpringIndizes = 2 * forcedNodes -1

                # print("Indizes")
                # print(xSpringIndizes)
                # print(ySpringIndizes)

                SpringC = self.Inp.springLocationData[:, 2].reshape(-1, 1)


                # Assign forces to degrees of freedom
                for i in range(0,len(SpringC),1):

                    if self.Inp.SpringDirections[i]  == "x":
                        Kp[xSpringIndizes[i]][xSpringIndizes[i]] += SpringC[i]
                        self.Kspring_diag[xSpringIndizes[i]] += SpringC[i]
                    elif self.Inp.SpringDirections[i] == "y":
                        Kp[ySpringIndizes[i]][ySpringIndizes[i]] += SpringC[i]
                        self.Kspring_diag[ySpringIndizes[i]] += SpringC[i]
            except:
                pass

        # Reduce to structure stiffness matrix by deleting rows and columns for restrained DoF
        if (len(self.Inp.restrainedIndex)>0):
            # print("RESTRAINED INDEX")
            # print(self.Inp.restrainedIndex)
            Ks = np.delete(Kp, self.Inp.restrainedIndex, 0)  # Delete rows
            Ks = np.delete(Ks, self.Inp.restrainedIndex, 1)  # Delete columns
        else:
            Ks = Kp
        

        Ks = np.matrix(
            Ks
        )  # Convert Ks from numpy.ndarray to numpy.matrix to use build in inverter function

       


        return Ks

    def solveDisplacements(self, Ks, F_inequilibrium):
        """
        Standard solving for structural displacements
        """

        forceVectorRed = copy.copy(
            F_inequilibrium
        )  # Make a copy of forceVector so the copy can be edited, leaving the original unchanged
        if (len(self.Inp.restrainedIndex)>0):
            forceVectorRed = np.delete(
                forceVectorRed, self.Inp.restrainedIndex, 0
            )  # Delete rows corresponding to restrained DoF
        else:
            forceVectorRed =  forceVectorRed

        #U = Ks.I * forceVectorRed
        U = np.linalg.solve(Ks, forceVectorRed)


        # Build the global displacement vector inclusing zeros as restrained degrees of freedom
        UG = np.zeros(
            self.Inp.nDoF
        )  # Initialise an array to hold the global displacement vector
        c = 0  # Initialise a counter to track how many restraints have been imposed
        for i in np.arange(self.Inp.nDoF):
            if i in self.Inp.restrainedIndex:
                # Impose zero displacement
                UG[i] = 0
            else:
                # Assign actual displacement
                UG[i] = U[c]
                c = c + 1

        UG = np.array([UG]).T

        return UG

    def SelfweigthLoadVector(self):

        if(self.swt):
            self.SW_at_supports = np.empty((0,2))
            for n, mbr in enumerate(self.Inp.members):  
                node_i = mbr[0] #Node number for node i of this member
                node_j = mbr[1] #Node number for node j of this member
                length = self.lengths[n]     
                sw = length*self.Mat.gamma[n] #(N) Self-weight of the member
                F_node = sw/2   #(N) Self-weight distributed into each node  
                # print("FNODE")
                # print(F_node)
                iy = 2*node_i-1 #index of y-DoF for node i
                jy = 2*node_j-1 #index of y-DoF for node j         
                self.forceVector[iy] = self.forceVector[iy] -F_node
                self.forceVector[jy] = self.forceVector[jy]  -F_node  
                
                #Check if SW needs to be directly added to supports (if elements connect to supports)
                if(iy+1 in self.Inp.restrainedDoF):
                    supportSW = np.array([iy, F_node])
                    self.SW_at_supports = np.append(self.SW_at_supports, [supportSW], axis=0) #Store y-DoF at support and force to be added
                if(jy+1 in self.Inp.restrainedDoF):
                    supportSW = np.array([jy, F_node])
                    self.SW_at_supports = np.append(self.SW_at_supports, [supportSW], axis=0) #Store y-DoF at support and force to be added                    
            print(self.forceVector)
            print(len(self.forceVector))
        else:
            pass

    def updateInternalForceSystem(self, UG):
        """
        Build internal force vector F_int (global DoF order) for the CURRENT increment UG.

        - Cable elements ('c'): nonlinear (deformed geometry, AA-matrix, etc.) -> uses self.TMs[n]
        - Bar elements   ('b'): linear Theorie I. Ordnung -> uses constant direction cosines from BarLinear

        IMPORTANT:
        - Pretension P0 is already handled separately via self.F_pre / self.F_inc initial column.
        Therefore for BAR internal force increment we DO NOT add P0 again here (avoid double count).
        """

        F_int = np.zeros((self.Inp.nDoF, 1), dtype=float)

        for n, mbr in enumerate(self.Inp.members):
            node_i, node_j = int(mbr[0]), int(mbr[1])

            # global DoF indices for this member
            ia = 2 * node_i - 2
            ib = 2 * node_i - 1
            ja = 2 * node_j - 2
            jb = 2 * node_j - 1

            # -------------------------
            # CABLE (nonlinear)
            # -------------------------
            if self.memberType[n] == "c":
                # incremental displacements (global)
                d_ix = float(UG[ia, 0])
                d_iy = float(UG[ib, 0])
                d_jx = float(UG[ja, 0])
                d_jy = float(UG[jb, 0])

                # current transformation (computed for cumulative shape, stored in self.TMs)
                TM = self.TMs[n, :, :]  # shape (2,4)

                # local incremental displacements
                localDisp = TM @ np.array([[d_ix], [d_iy], [d_jx], [d_jy]], dtype=float)
                u = float(localDisp[0, 0])
                v = float(localDisp[1, 0])

                # extension from nonlinear geometry
                Lo = float(self.lengths[n])
                e = math.sqrt((Lo + u) ** 2 + v**2) - Lo

                # AA matrix
                denom = (Lo + e)
                if abs(denom) < 1e-14:
                    # numerical guard (should not really happen)
                    continue

                a1 = (Lo + u) / denom
                a2 = v / denom
                AA = np.array([[a1, a2]], dtype=float)  # (1,2)

                # axial load increment (no P0 here; P0 was handled via initPretension)
                P = (float(self.Mat.E[n]) * float(self.Mat.A[n]) / Lo) * e

                # back to global nodal forces (4x1)
                F_global = (TM.T @ AA.T) * P  # (4,1)

                F_int[ia, 0] += float(F_global[0, 0])
                F_int[ib, 0] += float(F_global[1, 0])
                F_int[ja, 0] += float(F_global[2, 0])
                F_int[jb, 0] += float(F_global[3, 0])

            # -------------------------
            # BAR (linear, Theorie I. Ordnung)
            # -------------------------
            else:
                # element axial force increment from linear truss theory:
                # N = (EA/L) * [-c -s c s] * u_e
                f_e = self.BarLinear.internal_nodal_forces_global(
                    n,
                    UG,
                    self.Mat.E,
                    self.Mat.A,
                    P0=None,  # do NOT add P0 here (already in self.F_pre)
                )  # shape (4,1)

                F_int[ia, 0] += float(f_e[0, 0])
                F_int[ib, 0] += float(f_e[1, 0])
                F_int[ja, 0] += float(f_e[2, 0])
                F_int[jb, 0] += float(f_e[3, 0])

        return F_int

    def testForConvergence(self, it, threshold, F_ineq):
        """
        Test if structure has converged by comparing the maximum force in the equilibrium
        force vector against a threshold for the simulation.
        """
        notConverged = True  # Initialise the convergence flag
        maxIneq = 0
        if it > 0:
            maxIneq = np.max(abs(F_ineq[self.Inp.freeDoF]))
            
            if maxIneq < threshold:
                notConverged = False

        return notConverged, maxIneq

    def calculateMbrForces(self, UG):
        """
        Calculates the member forces based on change in length of each member
        Takes in the cumulative global displacement vector as UG
        """

        mbrForces = np.zeros(
            len(self.Inp.members)
        )  # Initialise a container to hold axial forces

        for n, mbr in enumerate(self.Inp.members):
            node_i = mbr[0]  # Node number for node i of this member
            node_j = mbr[1]  # Node number for node j of this member

            # Index of DoF for this member
            ia = 2 * node_i - 2  # horizontal DoF at node i of this member
            ib = 2 * node_i - 1  # vertical DoF at node i of this member
            ja = 2 * node_j - 2  # horizontal DoF at node j of this member
            jb = 2 * node_j - 1  # vertical DoF at node j of this member

            # New positions = initial pos + cum deflection
            ix = self.Inp.nodes[node_i - 1, 0] + UG[ia, 0]
            iy = self.Inp.nodes[node_i - 1, 1] + UG[ib, 0]
            jx = self.Inp.nodes[node_j - 1, 0] + UG[ja, 0]
            jy = self.Inp.nodes[node_j - 1, 1] + UG[jb, 0]

            dx = jx - ix  # x-component of vector along member
            dy = jy - iy  # y-component of vector along member
            newLength = math.sqrt(
                dx**2 + dy**2
            )  # Magnitude of vector (length of member)

            deltaL = newLength - self.lengths[n]  # Change in length
            force = (
                self.Mat.P0[n]
                + deltaL * self.Mat.E[n] * self.Mat.A[n] / self.lengths[n]
            )  # Axial force due to change in length and any pre-tension
            mbrForces[n] = force  # Store member force

        return mbrForces

    def AdditionalSupportForce(self):
        self.reactionsFlag = False #Initialise a flag so we can plot a message re. reactions when necessary
        if(self.swt):
            if self.SW_at_supports.size>0:   
                self.reactionsFlag = True
                for SW in self.SW_at_supports:        
                    index = int(SW[0]) #Index of the global force vector 'FG' to update        
                    self.FI_FINAL[index,:] = self.FI_FINAL[index,:] + SW[1] #Add nodal SW force directly to FG 

    def MainConvergenceLoop(self):
        i = 0  # Initialise an iteration counter (zeros out for each load increment)
        inc = 0  # Initialise load increment counter
        notConverged = True  # Initialise convergence flag

        # Init kspring-diagonal for the first iteration
        # It's overwriten in the generation of the stiffness matrix 
        # in each loop. Here just for the first run, where the stiffness matrix isn't initialized
        self.Kspring_diag = np.zeros(self.Inp.nDoF, dtype=float)

    
        self.forceIncrement = (
            self.forceVector / self.nForceIncrements
        )  # Determine the force increment for each convergence test
        self.maxForce = (
            self.forceVector
        )  # Define a vector to store the total external force applied
        self.forceVector = (
            self.forceIncrement
        )  # Initialise the forceVector to the first increment of load

        # print("Force vector")
        # print(self.forceVector)


        while notConverged and i < 10000:

            # Calculate the cumulative internal forces Fi_total = Fa + Fb + Fc + ...
            Fi_total = np.matrix(
                np.sum(self.F_inc, axis=1)
            ).T  # Sum across columns of F_inc

            # Calculate the cumulative incremental displacements UG_total = Xa + Xb + Xc + ...
            UG_total = np.matrix(
                np.sum(self.UG_inc, axis=1)
            ).T  # Sum across columns of UG_inc

            # Inequilibrium force vector used in this iteration F_EXT - Fi_total or externalForces - (cumulative) InternalForceSystem
            
            # add spring forces to internal force balance
            # (springs are in K, so their resisting forces must appear in equilibrium)
            F_spring = self.Kspring_diag.reshape(-1, 1) * np.asarray(UG_total, dtype=float)


            self.F_inequilibrium = self.forceVector - Fi_total - F_spring

            # Build the structure stiffness matrix based on current position (using cumulative displacements)
            Ks = self.buildStructureStiffnessMatrix(UG_total,self.TMs)

            # Solve for global (incremental) displacement vector [Xn] for this iteration
            self.UG = self.solveDisplacements(Ks, self.F_inequilibrium)

            # Calculate a new transformation matrix for each member based on cum disp up to previous iteration
            self.TMs = self.calculateTransMatrices(UG_total)

            # if i == 0:
            #     print(self.TMs)

            # Calculate the internal force system based on new incremental displacements, [Fn]
            F_int = self.updateInternalForceSystem(self.UG)

            # Save incremental displacements and internal forces for this iteration
            self.UG_inc = np.append(self.UG_inc, self.UG, axis=1)
            self.F_inc = np.append(self.F_inc, F_int, axis=1)

            # Test for convergence
            notConverged, maxIneq = self.testForConvergence(
                i, self.convThreshold, self.F_inequilibrium
            )

            i += 1

            # If system has converged, save converged displacements, forces and increment external loading
            if not notConverged:
                self.hasSlackElements = False #Initialise a flag to indicate presence of new slack elements
                mbrForces = self.calculateMbrForces(UG_total) #Calculate member forces based on current set of displacements
                
                #Test for compression in cable elements if designated number of converged increments reached

                if inc > self.checkSlackAfter:
                    for m, mbr in enumerate(self.Inp.members):            
                        if self.memberType[m] == 'c' and mbrForces[m]<0:
                            print(f'Compression in cable element from from nodes {mbr[0]} to {mbr[1]}')
                            self.hasSlackElements = True #Switch slack elements flag
                            self.Mat.A[m] = 0 #Eliminate member stiffness by seting its cross-sectional area to zero


                
                print(
                    f"System has converged for load increment {inc} after {i-1} iterations"
                )
                

                self.UG_FINAL = np.append(
                    self.UG_FINAL, UG_total, axis=1
                )  # Add the converged displacement record
                self.UG_inc = np.empty(
                    [self.Inp.nDoF, 0]
                )  # Zero out the record of incremental displacements for the next load increment
                self.UG_inc = np.array(
                    np.append(self.UG_inc, UG_total, axis=1)
                )  # Add the initial displacement record for next load increment (manually cast as ndarray instead of matrix)

                self.FI_FINAL = np.append(
                    self.FI_FINAL, Fi_total, axis=1
                )  # Add the converged force record
                self.F_inc = np.empty(
                    [self.Inp.nDoF, 0]
                )  # Zero out the record of incremental forces for the next load increment
                self.F_inc = np.array(
                    np.append(self.F_inc, Fi_total, axis=1)
                )  # Add the initial force record for next load increment (manually cast as ndarray instead of matrix)


                self.mbrForces = self.calculateMbrForces(
                    self.UG_FINAL[:, -1]
                )  # Calculate the member forces based on change in mbr length
                self.MBRFORCES = np.append(
                    self.MBRFORCES, np.matrix(self.mbrForces).T, axis=1
                )  # Add the converged axial forces record

                self.EXTFORCES = np.append(
                    self.EXTFORCES, self.forceVector, axis=1
                )  # Add the external force vector for this load increment

                # Test if all external loading has been applied
                if abs(sum(self.forceVector).item()) < abs(sum(self.maxForce).item()):
                    i = 0  # Reset counter for next load increment
                    inc += 1
                    self.forceVector = (
                        self.forceVector + self.forceIncrement
                    )  # Increment the applied load
                    notConverged = (
                        True  # Reset notConverged flag for next load increment
                    )

        self.AdditionalSupportForce()

    def SolveLinear_NoIteration(self, treat_cables_as_bars=True):
        """
        Linear solve (no iteration, no deformed geometry):
            K * u = F

        - Builds global stiffness matrix once from undeformed geometry.
        - Solves once for displacements.
        - Computes reactions and member forces.

        treat_cables_as_bars:
            True  -> use linear bar stiffness also for members typed 'c'
            False -> raise error if cables exist (strict linear truss only)
        """

        # ---------------------------------------------------------
        # 1) Build external load vector (global) once
        # ---------------------------------------------------------
        self.forceVector = np.zeros((len(self.Inp.nodes) * 2, 1), dtype=float)
        self.AddPointLoadsGlobal()
        self.calculateInitialLengths()
        self.SelfweigthLoadVector()   # only acts if self.swt=True

        F = self.forceVector.copy()   # global full vector (nDoF,1)

        # ---------------------------------------------------------
        # 2) Build global stiffness Kp (full) from undeformed geometry
        # ---------------------------------------------------------
        nDoF = self.Inp.nDoF
        Kp = np.zeros((nDoF, nDoF), dtype=float)

        # Springs: store diagonal like before
        self.Kspring_diag = np.zeros(nDoF, dtype=float)

        for n, mbr in enumerate(self.Inp.members):
            node_i, node_j = int(mbr[0]), int(mbr[1])

            if (self.memberType[n] == "c") and (not treat_cables_as_bars):
                raise ValueError(
                    f"Member {n+1} (nodes {node_i}-{node_j}) is a cable. "
                    "Set treat_cables_as_bars=True or remove cable members for strict linear solve."
                )

            # Use linear bar element stiffness (undeformed)
            K11, K12, K21, K22 = self.BarLinear.buildElementStiffnessMatrix(
                n, None, None, None, self.Mat.P0, self.Mat.E, self.Mat.A
            )

            ia = 2 * node_i - 2
            ib = 2 * node_i - 1
            ja = 2 * node_j - 2
            jb = 2 * node_j - 1

            Kp[ia:ib+1, ia:ib+1] += K11
            Kp[ia:ib+1, ja:jb+1] += K12
            Kp[ja:jb+1, ia:ib+1] += K21
            Kp[ja:jb+1, ja:jb+1] += K22

        # ---------------------------------------------------------
        # 3) Add springs (same logic as your iterative build)
        # ---------------------------------------------------------
        if len(self.Inp.springLocationData) > 0:
            try:
                forcedNodes = self.Inp.springLocationData[:, 1].astype(int)
                xIdx = 2 * forcedNodes - 2
                yIdx = 2 * forcedNodes - 1
                SpringC = self.Inp.springLocationData[:, 2].reshape(-1)

                for i in range(len(SpringC)):
                    c = float(SpringC[i])
                    if self.Inp.SpringDirections[i] == "x":
                        Kp[xIdx[i], xIdx[i]] += c
                        self.Kspring_diag[xIdx[i]] += c
                    elif self.Inp.SpringDirections[i] == "y":
                        Kp[yIdx[i], yIdx[i]] += c
                        self.Kspring_diag[yIdx[i]] += c
            except:
                pass

        # ---------------------------------------------------------
        # 4) Reduce and solve
        # ---------------------------------------------------------
        if len(self.Inp.restrainedIndex) > 0:
            free = np.array([i for i in range(nDoF) if i not in self.Inp.restrainedIndex], dtype=int)
        else:
            free = np.arange(nDoF, dtype=int)

        Kff = Kp[np.ix_(free, free)]
        Ff  = F[free, :]

        uf = np.linalg.solve(Kff, Ff)

        # build full displacement vector u (restrained are 0)
        u = np.zeros((nDoF, 1), dtype=float)
        u[free, 0] = uf[:, 0]

        # Save like your usual output containers (single step)
        self.UG_FINAL = u.copy()
        self.EXTFORCES = F.copy()

        # ---------------------------------------------------------
        # 5) Reactions (full)
        # ---------------------------------------------------------
        R = (Kp @ u) - F  # includes spring reactions etc.
        self.FI_FINAL = (Kp @ u)  # "internal nodal forces" equivalent
        self.Reactions = R

        # ---------------------------------------------------------
        # 6) Member forces (linear)
        #    include pretension if you want: P0 added in axial_force(..., P0=self.Mat.P0)
        # ---------------------------------------------------------
        mbrN = np.zeros((len(self.Inp.members), 1), dtype=float)
        for n in range(len(self.Inp.members)):
            # linear axial force (tension +)
            N = self.BarLinear.axial_force(n, u, self.Mat.E, self.Mat.A, P0=self.Mat.P0)
            mbrN[n, 0] = N

        self.MBRFORCES = mbrN.copy()

        return u, R, mbrN



    def Summarize(self):
        #Generate output statements
        print(f"OUTSTANDING FORCE IMBALANCE")
        for i in np.arange(0,self.Inp.nDoF):     
            if i not in self.Inp.restrainedIndex:
                print(f"Remaining force imbalance at DoF {i} is {round(self.F_inequilibrium[i,0]/1000,3)} kN")  

        maxInequality = round(max(abs(self.F_inequilibrium[self.Inp.freeDoF,0])).item()/1000,3)
        print(f"(max = {maxInequality} kN)")

        print("")
        print("REACTIONS")

        f_int = self.FI_FINAL[:,-1]
        for i in np.arange(0,len(self.Inp.restrainedIndex)):           
            index = self.Inp.restrainedIndex[i]
            print(f"Reaction at DoF {index+1}: {round(f_int[index].item()/1000,2)} kN")

        # last converged displacement vector
        u_last = np.asarray(self.UG_FINAL[:, -1], dtype=float).reshape(-1, 1)   # (nDoF,1)

        # spring stiffness (diagonal) as vector
        k = np.asarray(self.Kspring_diag, dtype=float).reshape(-1, 1)           # (nDoF,1)

        # elementwise spring forces
        f_springs = k * u_last                                                  # (nDoF,1)


        try:
            # Federkräfte (letzter Lastschritt)
            f = np.asarray(f_springs, dtype=float).flatten()

            springno = self.Inp.springLocationData[:, 0].astype(int)
            forcedNodes = self.Inp.springLocationData[:, 1].astype(int)
            dirs = np.asarray(self.Inp.SpringDirections)

            print("\nSPRING FORCES (per spring):")
            for no, node, d in zip(springno, forcedNodes, dirs):
                d = str(d).strip().lower().replace('"', '')

                if d == "x":
                    dof = 2 * node - 2
                elif d in ("y", "z"):
                    dof = 2 * node - 1
                else:
                    raise ValueError(f"Unknown spring direction: {d}")

                print(
                    f"Spring {no} | Node {node} | Dir {d} | "
                    f"DoF {dof} | Force = {f[dof]/1000:.2f} kN"
                )

        except Exception as e:
            print("No springs in the system")
            # optional:
            # print(e)

        print("")   
        print("MEMBER FORCES (incl. any pre-tension)")    
        for n, mbr in enumerate(self.Inp.members):    
            print(f"Force in member {n+1} (nodes {mbr[0]} to {mbr[1]}) is {round(self.mbrForces[n]/1000,2)} kN")

        print("")   
        print("NODAL DISPLACEMENTS") 
        ug = self.UG_FINAL[:,-1]
        for n, node in enumerate(self.Inp.nodes):    
            ix = 2*(n+1)-2 #horizontal DoF for this node
            iy = 2*(n+1)-1 #vertical DoF for this node
            
            ux = round(ug[ix,0],5) #Horizontal nodal displacement
            uy = round(ug[iy,0],5) #Vertical nodal displacement
            print(f"Node {n+1}: Ux = {ux} m, Uy = {uy} m")

    def SummarizeLinear(self):
        print("LINEAR SOLVE (NO ITERATION)")

        # -------------------------------------------------
        # REACTIONS
        # -------------------------------------------------
        print("\nREACTIONS (at restrained DoF):")
        for idx in self.Inp.restrainedIndex:
            print(f"DoF {idx+1}: R = {self.Reactions[idx,0]/1000:.2f} kN")

        # -------------------------------------------------
        # MEMBER FORCES
        # -------------------------------------------------
        print("\nMEMBER FORCES (incl. P0):")
        for n, mbr in enumerate(self.Inp.members):
            print(
                f"Member {n+1} (nodes {mbr[0]}-{mbr[1]}): "
                f"N = {self.MBRFORCES[n,0]/1000:.2f} kN"
            )

        # -------------------------------------------------
        # NODAL DISPLACEMENTS
        # -------------------------------------------------
        print("\nNODAL DISPLACEMENTS:")
        for n in range(len(self.Inp.nodes)):
            ix = 2 * (n + 1) - 2
            iy = 2 * (n + 1) - 1
            print(
                f"Node {n+1}: "
                f"Ux = {self.UG_FINAL[ix,0]:.6e} m, "
                f"Uy = {self.UG_FINAL[iy,0]:.6e} m"
            )

        # -------------------------------------------------
        # SPRING FORCES
        # -------------------------------------------------
        if len(self.Inp.springLocationData) == 0:
            print("\nNO SPRINGS IN SYSTEM")
            return

        print("\nSPRING FORCES:")

        # displacement vector
        u = self.UG_FINAL.reshape(-1, 1)

        # diagonal spring stiffness vector
        kdiag = self.Kspring_diag.reshape(-1, 1)

        # spring forces per DoF
        f_spring = kdiag * u

        try:
            spring_no  = self.Inp.springLocationData[:, 0].astype(int)
            nodes      = self.Inp.springLocationData[:, 1].astype(int)
            k_values   = self.Inp.springLocationData[:, 2].astype(float)
            directions = np.asarray(self.Inp.SpringDirections)

            for no, node, k_i, d in zip(spring_no, nodes, k_values, directions):
                d = str(d).strip().lower().replace('"', '')

                if d == "x":
                    dof = 2 * node - 2
                elif d == "y":
                    dof = 2 * node - 1
                else:
                    raise ValueError(f"Unknown spring direction: {d}")

                print(
                    f"Spring {no} | Node {node} | Dir {d} | "
                    f"k = {k_i:.3e} N/m | "
                    f"u = {u[dof,0]:.6e} m | "
                    f"F = {f_spring[dof,0]/1000:.2f} kN"
                )

        except Exception as e:
            print("Error while printing spring forces")
            # print(e)




class BeamSubSection:

    def __init__(self, moment_ed, normal_force_ed, shear_force_ed, effective_height, effective_height_pressure, elasticity_modulus_steel):
        self.moment_ed = moment_ed
        self.normal_force_ed = normal_force_ed
        self.shear_force_ed = shear_force_ed
        self.effective_height = effective_height
        self.effective_height_pressure = effective_height_pressure
        self.elasticity_modulus_steel = elasticity_modulus_steel



from KIB_LAP.Betonbau import BeamRectangular
from KIB_LAP.Betonbau import BeamSubSection

import numpy as np
import matplotlib.pyplot as plt
from matplotlib import cm
import pandas as pd
import time
import os


Beam = BeamRectangular(1,1.00,0.60)

moment_ed = 0.20
normal_force_ed = 0.20
shear_force_ed  = 0.30
effective_height = 0.65
effective_height_pressure = 0.08
elasticity_modulus_steel = 200000 

BeamSection = BeamSubSection(moment_ed, normal_force_ed,shear_force_ed, effective_height,effective_height_pressure, elasticity_modulus_steel)


Beam.calculate_beam_section_without_shearreinforcement(BeamSection,30*0.01**2,35,500)


print(Beam.v_rdcs * 1000)
print(Beam.v_rdcmin * 1000)


class cross_section_polygon:
    def __init__(self, y, z):
        """_summary_
        Calculation of cross section properties for polygons. The coordinates need to be inserted clockwise  \n
        for positive areas, and anti-clockwise for wholes. \n
        Args:
            y (_type_): _description_
            z (_type_): _description_
        """
        self.y_restart = [float(y_f) for y_f in y]
        self.z_restart = [float(z_f) for z_f in z]
        self.len_y = len(self.y_restart)
        A_Test = 0
        for row in range(0, self.len_y - 1, 1):
            A_Test += 0.5 * (
                self.z_restart[row + 1] * self.y_restart[row]
                - self.z_restart[row] * self.y_restart[row + 1]
            )

        if A_Test >= 0:
            self.y = [float(y_f) for y_f in y]
            self.y.append(
                self.y[0]
            )  # Last element is necessary for loop over polygon (left rotation direction)
            self.z = [float(z_f) for z_f in z]
            self.z.append(self.z[0])
        elif A_Test < 0:
            self.y = [float(y_f) for y_f in y]
            self.y.append(self.y[0])
            self.y = self.y[::-1]
            self.z = [float(z_f) for z_f in z]
            self.z.append(self.z[0])
            self.z = self.z[::-1]

        print(self.y)
        print(self.z)

        print(self.len_y)
        self.A = 0
        self.S_y = 0
        self.S_z = 0
        self.I_yy = 0
        self.I_zz = 0
        self.I_yz = 0
        self.ys = 0
        self.zs = 0

        self.list_names = []
        self.list_unit = []
        self.list_value = []

        self.section_props()

    def section_props(self):
        print("Section_Props is accessed")
        self.list_names = []
        self.list_unit = []
        self.list_value = []
        for row in range(0, self.len_y - 1, 1):
            self.A += 0.5 * (
                self.z[row + 1] * self.y[row] - self.z[row] * self.y[row + 1]
            )
            print(self.A)
            self.S_y += (
                1
                / 6
                * (
                    (self.z[row] + self.z[row + 1])
                    * (self.z[row + 1] * self.y[row] - self.z[row] * self.y[row + 1])
                )
            )
            self.S_z += (
                1
                / 6
                * (
                    (self.y[row] + self.y[row + 1])
                    * (self.z[row + 1] * self.y[row] - self.z[row] * self.y[row + 1])
                )
            )
            self.I_yy += (
                1
                / 12
                * (
                    (
                        self.z[row + 1] ** 2
                        + (self.z[row] + self.z[row + 1]) * self.z[row]
                    )
                    * (self.z[row + 1] * self.y[row] - self.z[row] * self.y[row + 1])
                )
            )
            self.I_zz += (
                1
                / 12
                * (
                    (
                        self.y[row + 1] ** 2
                        + (self.y[row] + self.y[row + 1]) * self.y[row]
                    )
                    * (self.z[row + 1] * self.y[row] - self.z[row] * self.y[row + 1])
                )
            )
            self.I_yz += (
                1
                / 12
                * (
                    1 / 2 * self.y[row + 1] ** 2 * self.z[row] ** 2
                    - 1 / 2 * self.y[row] ** 2 * self.z[row + 1] ** 2
                    - (self.z[row + 1] * self.y[row] - self.z[row] * self.y[row + 1])
                    * (self.y[row] * self.z[row] + self.y[row + 1] * self.z[row + 1])
                )
            )

        self.ys = self.S_z / self.A
        self.zs = self.S_y / self.A

        self.I_yy = self.I_yy - self.zs**2 * self.A
        self.I_zz = self.I_zz - self.ys**2 * self.A
        self.I_yz = self.I_yz + self.ys * self.zs * self.A

        

        print("Area", self.A)

        print("First moment of area in y-direction", self.S_y)

        print("Second moment of area in y-direction", self.I_yy)

        self.list_names.append("Area")
        self.list_names.append("First moment of area in y-direction Sy")
        self.list_names.append("First moment of area in z-direction Sz")
        self.list_names.append("Second moment of area about y-Axis Iyy")
        self.list_names.append("Second moment of area about z-Axis Izz")

        self.list_unit.append("m**2")
        self.list_unit.append("m**3")
        self.list_unit.append("m**3")
        self.list_unit.append("m**4")
        self.list_unit.append("m**4")

        self.list_value.append(self.A)
        self.list_value.append(self.S_y)
        self.list_value.append(self.S_z)
        self.list_value.append(self.I_yy)
        self.list_value.append(self.I_zz)

        return self.list_names, self.list_value, self.list_unit
    



# _y = [0,0,2,2]
# _z = [0,1,1,0]


# Test = cross_section_polygon(_y,_z)


import sympy as sp
import numpy as np
import math


class ElemStema:
    def __init__(self):

        # Initialisierung der Steifigkeitsmatrix
        self.Ke = np.zeros((14, 14))

    def TransformationMatrix(self, posI, posJ):
        Pi = np.asarray(posI, dtype=float).reshape(
            3,
        )
        Pj = np.asarray(posJ, dtype=float).reshape(
            3,
        )

        v = Pj - Pi
        L = np.linalg.norm(v)

        T = np.zeros((14, 14), dtype=float)

        # identische Knoten
        if L < 1e-12:
            np.fill_diagonal(T, 1.0)
            return T

        # lokale x-Achse
        ex = v / L

        # robuste Referenzachse wählen (nicht parallel zu ex)
        g = np.array([0.0, 1.0, 0.0], dtype=float)  # bevorzugt global y
        if abs(np.dot(ex, g)) > 0.95:
            g = np.array([0.0, 0.0, 1.0], dtype=float)  # sonst global z
            if abs(np.dot(ex, g)) > 0.95:
                g = np.array([1.0, 0.0, 0.0], dtype=float)  # zur Not global x

        # --- Stabiler Aufbau: erst ey als Projektion von g orthogonal zu ex ---
        ey = g - np.dot(g, ex) * ex
        ney = np.linalg.norm(ey)
        if ney < 1e-12:
            raise ValueError("Degeneriertes lokales Achsensystem (ey ~ 0).")
        ey = ey / ney

        # --- Rechtshändiges System erzwingen ---
        ez = np.cross(ex, ey)
        nez = np.linalg.norm(ez)
        if nez < 1e-12:
            raise ValueError("Degeneriertes lokales Achsensystem (ez ~ 0).")
        ez = ez / nez

        # Optional: Flip-Schutz (ey soll möglichst in Richtung g zeigen)
        # verhindert, dass bei ähnlichen Elementen ey/ez um 180° springen
        if np.dot(ey, g) < 0.0:
            ey *= -1.0
            ez *= -1.0

        # R: local -> global (Spalten sind ex, ey, ez)
        R = np.column_stack((ex, ey, ez))

        # ---- 6x6 Transformationsblock passend zu deiner DOF-Reihenfolge ----
        Ti = np.zeros((6, 6), dtype=float)

        # translations: [ux, uy, uz] sitzen auf [0,1,3]
        trans = [0, 1, 3]
        # rotations: [rx, ry, rz] sitzen auf [5,4,2]
        rot = [5, 4, 2]

        Ti[np.ix_(trans, trans)] = R
        Ti[np.ix_(rot, rot)] = R

        # In 14x14 einbauen (Knoten a: 0..6, Knoten b: 7..13)
        T[0:6, 0:6] = Ti
        T[7:13, 7:13] = Ti

        # Warping
        T[6, 6] = 1.0
        T[13, 13] = 1.0

        # Debug-Sanity
        if not np.isfinite(T).all():
            print("TM NaN/Inf:", posI, posJ)
            print(T)
            raise ValueError("TM enthält NaN/Inf")

        return T

    def insert_elements(
        self, S, E, G, A, I_y, I_z, I_omega, I_T, cv, z1, cw, z2, c_thet, l
    ):
        """
        Element-Stiffness-Matrix:
        Na
        Vya
        Mza
        Vza
        Mya
        Mxa
        Mwa
        Nb
        Vyb
        Mzb
        Vzb
        Myb
        Mxb
        Mwb
        """
        self.Ke[:, :] = 0.0
        self.S = S  # Stiffness of shear field

        self.E = E  # Material stiffness of the beam
        self.G = G

        self.A = A
        self.I_y = I_y
        self.I_z = I_z

        self.I_omega = I_omega
        self.I_T = I_T

        self.cv = cv
        self.z1 = z1
        self.cw = cw
        self.z2 = z2

        self.c_thet = c_thet

        self.l = l
        # Matrixeinträge gemäß Tabelle definieren
        self.Ke[0, 0] = self.Ke[7, 7] = self.E * self.A / self.l
        self.Ke[0, 7] = self.Ke[7, 0] = -self.E * self.A / self.l

        self.Ke[1, 1] = self.Ke[8, 8] = (
            12 * self.E * self.I_z / self.l**3
            + 13 / 35 * self.cv * self.l
            + 1.2 * self.S / self.l
        )
        self.Ke[1, 2] = (
            6 * self.E * self.I_z / self.l**2
            + 11 / 210 * self.cv * self.l**2
            + 0.1 * self.S
        )

        self.Ke[1, 5] = (
            13 / 35 * self.cv * self.l * self.z1 - 1.2 * self.S / self.l * self.z2
        )
        self.Ke[1, 6] = (
            -11 / 210 * self.cv * self.l**2 * self.z1 + 0.1 * self.S * self.z2
        )

        self.Ke[1, 8] = (
            -12 * self.E * self.I_z / self.l**3
            + 9 / 70 * self.cv * self.l
            - 1.2 * self.S / self.l
        )
        self.Ke[1, 9] = (
            6 * self.E * self.I_z / self.l**2
            - 13 / 420 * self.cv * self.l**2
            + 0.1 * self.S
        )

        self.Ke[1, 12] = (
            9 / 70 * self.cv * self.l * self.z1 + 1.2 * self.S / self.l * self.z2
        )
        self.Ke[1, 13] = 13 / 420 * self.cv * self.l**2 * z1 + 0.1 * self.S * self.z2

        self.Ke[2, 2] = (
            4 * self.E * self.I_z / self.l
            + 1 / 105 * self.cv * self.l**3
            + 2 / 15 * self.S * self.l
        )
        self.Ke[9, 9] = self.Ke[2, 2]

        self.Ke[2, 5] = 11 / 210 * self.cv * l**2 * self.z1 - 0.1 * self.S * self.z2
        self.Ke[2, 6] = (
            -1 / 105 * self.cv * self.l**3 * self.z1
            + 2 / 15 * self.S * self.l * self.z2
        )
        self.Ke[2, 8] = (
            -6 * self.E * self.I_z / self.l**2
            + 13 / 420 * self.cv * l**2
            - 0.1 * self.S
        )

        self.Ke[2, 9] = (
            2 * self.E * self.I_z / self.l
            - 1 / 140 * self.cv * self.l**3
            - 1 / 30 * self.S * self.l
        )
        self.Ke[2, 12] = (
            13 / 420 * self.cv * self.l**2 * self.z1 + 0.1 * self.S * self.z2
        )
        self.Ke[2, 13] = (
            1 / 140 * self.cv * self.l**3 * self.z1 - 1 / 30 * self.S * self.l * self.z2
        )

        self.Ke[3, 3] = 12 * self.E * self.I_y / self.l**3 + 13 / 35 * self.cw * self.l
        self.Ke[10, 10] = self.Ke[3, 3]

        self.Ke[3, 4] = (
            -6 * self.E * self.I_y / self.l**2 - 11 / 210 * self.cw * self.l**2
        )
        self.Ke[3, 10] = -12 * self.E * self.I_y / self.l**3 + 9 / 70 * self.cw * self.l
        self.Ke[3, 11] = (
            -6 * self.E * self.I_y / self.l**2 + 13 / 420 * self.cw * self.l**2
        )

        self.Ke[4, 4] = 4 * self.E * self.I_y / self.l + 1 / 105 * self.cw * self.l**3
        self.Ke[11, 11] = self.Ke[4, 4]

        self.Ke[4, 10] = (
            6 * self.E * self.I_y / self.l**2 - 13 / 420 * self.cw * self.l**2
        )
        self.Ke[4, 11] = 2 * self.E * self.I_y / self.l - 1 / 140 * self.cw * self.l**3

        self.Ke[5, 5] = self.Ke[12, 12] = (
            12 * self.E * self.I_omega / self.l**3
            + 1.2 * self.G * self.I_T / self.l
            + 13 / 35 * self.c_thet * self.l
            + 13 / 35 * self.cv * self.l * self.z1**2
            + 1.2 * self.S / self.l * self.z2**2
        )
        self.Ke[5, 6] = (
            -6 * self.E * self.I_omega / self.l**2
            - 0.1 * self.G * self.I_T
            - 11 / 210 * self.c_thet * self.l**2
            - 11 / 210 * self.cv * self.l**2 * self.z1**2
            - 0.1 * self.S * self.z2**2
        )
        self.Ke[5, 8] = 9 / 70 * self.cv * l * self.z1 + 1.2 * self.S / self.l * self.z2
        self.Ke[5, 9] = -13 / 420 * self.cv * self.l * self.z1 - 0.1 * self.S * self.z2
        self.Ke[5, 12] = (
            -12 * self.E * self.I_omega / self.l**3
            - 1.2 * self.G * self.I_T / self.l
            + 9 / 70 * self.c_thet * self.l
            + 9 / 70 * self.cv * self.l * self.z1**2
            - 1.2 * self.S / self.l * self.z2**2
        )
        self.Ke[5, 13] = (
            -6 * self.E * self.I_omega / self.l**2
            - 0.1 * self.G * self.I_T
            + 13 / 420 * self.c_thet * self.l**2
            + 13 / 420 * self.cv * self.l**2 * self.z1**2
            - 0.1 * self.S * self.z2**2
        )
        self.Ke[6, 6] = self.Ke[13, 13] = (
            4 * self.E * self.I_omega / self.l
            + 2 / 15 * self.G * self.I_T * self.l
            + 1 / 105 * self.c_thet * self.l**3
            + 1 / 105 * self.cv * self.l**3 * self.z1
            + 2 / 15 * self.S * self.l * self.z2**2
        )

        self.Ke[6, 8] = (
            -13 / 420 * self.cv * self.l**2 * self.z1 - 0.1 * self.S * self.z2
        )
        self.Ke[6, 9] = (
            1 / 140 * self.cv * self.l**3 * self.z1 - 1 / 30 * self.S * self.l * self.z2
        )
        self.Ke[6, 12] = (
            6 * self.E * self.I_omega / self.l**2
            + 0.1 * self.G * self.I_T
            - 13 / 420 * self.c_thet * self.l**2
            - 13 / 420 * self.cv * self.l**2 * self.z1**2
            + 0.1 * self.S * self.z2**2
        )
        self.Ke[6, 13] = (
            2 * self.E * self.I_omega / self.l
            - 1 / 30 * self.G * self.I_T * self.l
            - 1 / 140 * self.c_thet * self.l**3
            - 1 / 140 * self.cv * l**3 * self.z1**2
            - 1 / 30 * self.S * self.l * self.z2**2
        )
        self.Ke[8, 9] = (
            -6 * self.E * self.I_z / self.l**2
            - 11 / 210 * self.cv * self.l**2
            - 0.1 * self.S
        )
        self.Ke[8, 12] = (
            13 / 35 * self.cv * self.l * self.z1 - 1.2 * self.S / self.l * self.z2
        )
        self.Ke[8, 13] = (
            11 / 210 * self.cv * self.l**2 * self.z1 - 0.1 * self.S * self.z2
        )
        self.Ke[9, 12] = (
            -11 / 210 * self.cv * self.l**2 * self.z1 + 0.1 * self.S * self.z2
        )
        self.Ke[9, 13] = (
            -1 / 105 * self.cv * self.l**3 * z1 + 2 / 15 * self.S * self.l * self.z2
        )
        self.Ke[10, 11] = 6 * E * I_y / l**2 + 11 / 210 * cw * l**2
        self.Ke[12, 13] = (
            6 * E * I_omega / l**2
            + 0.1 * G * I_T
            + 11 / 210 * c_thet * l**2
            + 11 / 210 * cv * l**2 * z1**2
            + 0.1 * S * z2**2
        )

        # Elem Matrix is symmetrical

        for i in range(14):
            for j in range(i):
                self.Ke[i, j] = self.Ke[j, i]

        return self.Ke

    def print_elem_matrix(self):
        sp.pprint(self.Ke)

    def Kg_theory_II_order(
        self, L, N, My_a, My_b, Mz_a, Mz_b, Mr, qy, qz, yq, zq, yM, zM
    ):
        """
        Geometrische Steifigkeitsmatrix K_g (14x14) für Theorie II. Ordnung
        inkl. Biegedrillknicken (Vlasov-Element mit Warping), gemäß deiner Belegung.

        DOF-Reihenfolge (1..14) wie in deinem Element:
          1: Na
          2: Vya
          3: Mza
          4: Vza
          5: Mya
          6: Mxa
          7: Mwa
          8: Nb
          9: Vyb
         10: Mzb
         11: Vzb
         12: Myb
         13: Mxb
         14: Mwb
        """
        Kg = np.zeros((14, 14), dtype=float)

        mq = qy * (yq - yM) + qz * (zq - zM)

        def setg(i, j, val):
            # i,j in 1-based
            Kg[i - 1, j - 1] = val
            Kg[j - 1, i - 1] = val  # Symmetrie

        # --- Einträge gemäß Screenshot ---
        for i, j in [(2, 2), (4, 4), (9, 9), (11, 11)]:
            setg(i, j, 1.2 * N / L)

        for i, j in [(2, 3), (2, 10), (5, 11), (11, 12)]:
            setg(i, j, 0.1 * N)

        setg(
            2,
            6,
            -1.1 * My_a / L - 0.1 * My_b / L - qz * L * (9 / 140) + 1.2 * (zM / L) * N,
        )
        setg(2, 7, 0.1 * My_a + qz * (L**2) / 140 - 0.1 * zM * N)
        setg(2, 9, -1.2 * N / L)
        setg(4, 11, -1.2 * N / L)
        setg(
            2,
            13,
            0.1 * My_a / L + 1.1 * My_b / L + qz * L * (9 / 140) - 1.2 * (zM / L) * N,
        )
        setg(2, 14, 0.1 * My_b + qz * (L**2) / 140 + 0.1 * zM * N)

        for i, j in [(3, 3), (5, 5), (10, 10), (12, 12)]:
            setg(i, j, (2 / 15) * N * L)

        setg(3, 6, -0.9 * My_a - 0.2 * My_b - qz * (L**2) * (31 / 420) + 0.1 * zM * N)
        setg(3, 7, 0.1 * My_a + (1 / 30) * My_b + qz * (L**2) / 84 - (2 / 15) * zM * N)

        for i, j in [(3, 9), (4, 5), (4, 12), (9, 10)]:
            setg(i, j, -0.1 * N)

        for i, j in [(3, 10), (5, 12)]:
            setg(i, j, -(1 / 30) * N * L)

        setg(3, 13, -0.1 * My_a + 0.2 * My_b - qz * (L**2) / 105 + 0.1 * zM * N)
        setg(3, 14, -(1 / 30) * My_a * L - qz * (L**2) / 210 + (1 / 30) * L * zM * N)

        setg(4, 6, -1.1 * Mz_a / L - 0.1 * Mz_b / L - 1.2 * (yM / L) * N)
        setg(4, 7, 0.1 * Mz_a + 0.1 * yM * N)
        setg(4, 13, 0.1 * Mz_a / L + 1.1 * Mz_b / L + 1.2 * (yM / L) * N)
        setg(4, 14, 0.1 * Mz_b + 0.1 * yM * N)

        setg(5, 6, 0.9 * Mz_a + 0.2 * Mz_b + 0.1 * yM * N)
        setg(5, 7, -0.1 * Mz_a * L - (1 / 30) * Mz_b * L - (2 / 15) * L * yM * N)
        setg(5, 13, 0.1 * Mz_a - 0.2 * Mz_b - 0.1 * yM * N)
        setg(5, 14, (1 / 30) * Mz_a * L + (1 / 30) * L * yM * N)

        setg(6, 6, 1.2 * Mr / L + mq * (13 / 35) * L)
        setg(13, 13, 1.2 * Mr / L + mq * (13 / 35) * L)

        setg(6, 7, -0.1 * Mr - mq * (11 / 210) * (L**2))
        setg(
            6,
            9,
            1.1 * My_a / L + 0.1 * My_b / L + qz * L * (9 / 140) - 1.2 * (zM / L) * N,
        )
        setg(6, 10, -0.2 * My_a + 0.1 * My_b + qz * (L**2) / 105 + 0.1 * zM * N)
        setg(6, 11, 1.1 * Mz_a / L + 0.1 * Mz_b / L + 1.2 * (yM / L) * N)
        setg(6, 12, 0.2 * Mz_a - 0.1 * Mz_b + 0.1 * yM * N)
        setg(6, 13, -1.2 * Mr / L - mq * L * (9 / 70))
        setg(6, 14, -0.1 * Mr + mq * (13 / 420) * (L**2))

        setg(7, 7, (L / 7.5) * Mr + mq * (L**3) / 105)
        setg(14, 14, (L / 7.5) * Mr + mq * (L**3) / 105)

        setg(7, 9, -0.1 * My_a - qz * (L**2) / 140 + 0.1 * zM * N)
        setg(7, 10, -(1 / 30) * My_a * L - qz * (L**2) / 210 + (1 / 30) * L * zM * N)
        setg(7, 11, -0.1 * Mz_a - 0.1 * yM * N)
        setg(7, 12, (1 / 30) * Mz_b * L + (1 / 30) * L * yM * N)
        setg(7, 13, 0.1 * Mr - mq * (13 / 420) * (L**2))
        setg(7, 14, -(1 / 30) * Mr * L - mq * (L**3) / 140)

        setg(
            9,
            13,
            -0.1 * My_a / L - 1.1 * My_b / L - qz * L * (9 / 140) + 1.2 * (zM / L) * N,
        )
        setg(9, 14, -0.1 * My_b - qz * (L**2) / 140 + 0.1 * zM * N)

        setg(10, 13, 0.2 * My_a - 0.9 * My_b + qz * (L**2) * (31 / 420) - 0.1 * zM * N)
        setg(
            10,
            14,
            (1 / 30) * My_a * L
            + 0.1 * My_b * L
            + qz * (L**2) / 84
            - (2 / 15) * L * zM * N,
        )

        setg(11, 13, -0.1 * Mz_a / L - 1.1 * Mz_b / L - 1.2 * (yM / L) * N)
        setg(11, 14, -0.1 * Mz_b - 0.1 * yM * N)

        setg(12, 13, -0.2 * Mz_a - 0.9 * Mz_b - 0.1 * yM * N)
        setg(12, 14, -(1 / 30) * Mz_a * L - 0.1 * Mz_b * L - (2 / 15) * L * yM * N)

        setg(13, 14, 0.1 * Mr + mq * (11 / 210) * (L**2))

        return Kg

    def element_stiffness_theory_II(
        self,
        # Material-/Elementparameter für Ke:
        S,
        E,
        G,
        A,
        I_y,
        I_z,
        I_omega,
        I_T,
        cv,
        z1,
        cw,
        z2,
        c_thet,
        l,
        # Schnittgrößen/Lasten für Kg:
        N,
        My_a,
        My_b,
        Mz_a,
        Mz_b,
        Mr,
        qy,
        qz,
        yq,
        zq,
        yM,
        zM,
    ):
        """
        Liefert K_total = Ke + Kg (lokal).
        """
        Ke = self.insert_elements(
            S, E, G, A, I_y, I_z, I_omega, I_T, cv, z1, cw, z2, c_thet, l
        )
        Kg = self.Kg_theory_II_order(
            L=l,
            N=N,
            My_a=My_a,
            My_b=My_b,
            Mz_a=Mz_a,
            Mz_b=Mz_b,
            Mr=Mr,
            qy=qy,
            qz=qz,
            yq=yq,
            zq=zq,
            yM=yM,
            zM=zM,
        )
        return Ke + Kg, Ke, Kg
